#!/usr/bin/perl

# PLVote

#
# Written by Riad Wahby.  This software is
# hereby released into the public domain under
# the terms of the GNU public license.  This
# software may be freely modified and redistributed
# so long as this header remains in its entirety.
#
# 8/10/99 - rsw@phonon.net
#

use LWP::UserAgent;

$version = "1.0.2";

$ftpurl = "ftp://ftp.phonon.net/pub/plvote/";

$maxtries = 10000; # max tries to open a locked file before dying.
                   # 10000 is a decent number

$| = 1;

### Subprogram declarations

sub read_data;       # read in data from POST

sub check_data;      # check data POSTed

sub print_results;   # print the current results

sub process_vote;    # processes a single POSTed vote

sub print_comments;  # prints all comments

sub print_header;    # gets and dumps user-specified header or, if
                     # not specified, a generic header

sub print_footer;    # gets and dumps user-specified footer or, if
                     # not specified, a generic footer

sub http_dump;       # dumps the contents of a specified URL

sub touch;           # basic UNIX touch used by safeopen

sub safeopen;        # implements file locking--used for NT portability

sub safeclose;       # safely closes files

sub print_help;      # print the HTML help for this script

################
# MAIN SECTION #
#########################################
# A dispatch for the rest of the script #
#########################################

print "Content-type: text/html\n\n";

read_data();

unless (check_data)
{
    print_help();
    exit;
}

touch ($FORM{'comment_file'});
unless (-e $FORM{'vote_file'})
{
    my $newfile;
    foreach $number (@questions)
    {
	$newfile .= "$number:0:" . $FORM{"a$number"} . "\n";
    }
    open (VOTES, ">$FORM{'vote_file'}");
    print VOTES $newfile;
    close (VOTES);
}

if ($FORM{'vote_qnum'} && $FORM{"a" . $FORM{'vote_qnum'}})
{
    process_vote;
}

print_header;

print_results;

print_footer;

foreach my $file (keys %OPENFILES)
{
    close $file;
    unlink ($file . ".lock");
}

exit;

############
# End MAIN #
############

#############
# READ_DATA #
#############
sub read_data
{
#
# takes no arguments
#
# reads in POST data
#
# returns 1 on success, 0 otherwise
#
    my $buffer;
    read(STDIN, $buffer, $ENV{'CONTENT_LENGTH'});
    my @pairs = split(/&/, $buffer);
    foreach $pair (@pairs)
    {
	
# convert plus signs to spaces
	$pair =~ s/\+/ /g;
	
# convert encoded data to characters	
	$pair =~ s/%(..)/pack("c",hex($1))/ge;
	
# Finally, put the results in a hash, FORM
	my $loc=index($pair,"=");
	my $param=substr($pair, 0, $loc);
	$param = "\L$param\E";   # Convert to all lowercase
	my $value=substr($pair, $loc+1);
	$FORM{$param} .= $value;
	if ($param =~ /^a(\d+)$/)
	{
	    push @questions, $1;
	}
    }
    return 1;
}

##############
# CHECK_DATA #
##############
sub check_data
{
#
# takes no arguments
#
# checks the POST data
#
# returns 1 on success, 0
# if required fields are missing
#
    foreach my $required ("vote_file","comment_file","barsrc")
    {
	unless ($FORM{$required})
	{
	    return 0;
	}
    }
    my %defaults = ("bgcolor" => "#FFFFFF",
		    "vote_bgcolor" => \$FORM{'bgcolor'},
		    "comment_title_bgcolor" => "#00B4ED",
		    "comment_body_bgcolor" => \$FORM{'bgcolor'},
		    "name" => "Anonymous",
		    "title" => "Untitled Comment",
		    "graph_width" => 390,
		    "vote_title" => "Voting Results",
		    "comment_title" => "Comments",
		    "page_title" => "Voting Results",
		    "bar_height" => "25",
		    "comment_table_width" => "100%",
		    "comment_body_font" => "#000000",
		    "comment_title_font" => \$FORM{'comment_body_font'});
    foreach my $default (sort keys %defaults)
    {
	unless ($FORM{$default})
	{
	    if (ref $defaults{$default})
	    {
		$FORM{$default} = ${$defaults{$default}};
	    }
	    else
	    {
		$FORM{$default} = $defaults{$default};
	    }
	}
    }
    return 1;
}		 

#################
# PRINT_RESULTS #
#################
sub print_results
{
#
# takes no input
#
# prints the results of the vote so far
#
# always returns true
#
    print "<h3>$FORM{'vote_title'}</h3>\n";
    print "<table border=0 bgcolor=$FORM{'vote_bgcolor'} background=$FORM{'vote_background'} align=center>\n";
    my $votetotal = 0;
    if (safeopen (PRINTVOTES, "$FORM{'vote_file'}", "<"))
    {
	my %VOTES;
	my %QUESTS;
	while (<PRINTVOTES>)
	{
	    my ($qnum, $total, $question) = split (/:/, $_);
	    $VOTES{$qnum} = $total;
	    $QUESTS{$qnum} = $question;
	    $votetotal += $total;
	}
	safeclose (PRINTVOTES);
	foreach $key (sort keys %VOTES)
	{
	    my $votepercent = $VOTES{$key} / $votetotal;
	    my $votewidth = int ($votepercent * $FORM{'graph_width'});
	    $votepercent = sprintf "%4.3f", $votepercent;
	    chomp $votepercent;
	    $votepercent *= 100;
	    print <<END;
<tr>
<th align=left>$QUESTS{$key}:</th>
<td align=left>$VOTES{$key}</td>
<td align=left width=$FORM{'graph_width'}><img src="$FORM{'barsrc'}" width=$votewidth height="$FORM{'bar_height'}"></td>
<td align=left>$votepercent%</td>
</tr>
END
        }
	print "<tr><th align=right>Total: $votetotal Votes</th></tr></table>\n";
    }
    else
    {
	if ($FORM{'debug'})
	{
	    print "<strong>Unable to open votes file $FORM{'vote_file'} in PRINT_RESULTS</strong>\n";
	}
    }
    if ($FORM{'show_comments'})
    {
	print_comments();
    }
}
    
################
# PROCESS_VOTE #
################
sub process_vote
{
#
# takes no input
#
# processes a vote from a user POST
#
# always returns true
#
    if (safeopen (PROCESSVOTES, "$FORM{'vote_file'}", "<"))
    {
	my $voted = 0;
	my $newvote;
	while ($line = <PROCESSVOTES>)
	{
	    my ($qnum, $total, $question) = split (/:/, $line);
	    unless ($FORM{"a$qnum"})
	    {
		$FORM{"a$qnum"} = $question;
	    }
	    if ($FORM{'vote_qnum'} eq $qnum)
	    {
		$total++;
		$voted = 1;
	    }
	    $newvote .= "$qnum:$total:" . $FORM{"a$qnum"} . "\n";
	}
	close (PROCESSVOTES);
	unless ($voted)
	{
	    $newvote .= $FORM{'vote_qnum'} . ":1:" . $FORM{"a$FORM{'vote_qnum'}"} . "\n";
	}
	unlink ($FORM{'vote_file'});
	open (PROCESSVOTES, ">$FORM{'vote_file'}");
	print PROCESSVOTES $newvote;
	safeclose (PROCESSVOTES);
    }
    else
    {
	if ($FORM{'debug'})
	{
	    print "<strong>Unable to open votes file $FORM{'vote_file'} in PROCESS_VOTE</strong>\n";
	}
    }
    if ($FORM{'comment'})
    {
	if (safeopen (PROCESSCOMMENTS, "$FORM{'comment_file'}", ">>"))
	{
	    my $timestamp = scalar(localtime);
	    $FORM{'comment'} =~ s/\^//g;
	    print PROCESSCOMMENTS "$FORM{'name'}^$FORM{'email'}^$timestamp^$FORM{'title'}^$FORM{'comment'}\n";
	    safeclose (PROCESSCOMMENTS);
	}
	else
	{
	    if ($FORM{'debug'})
	    {
		print "<strong>Unable to open comments file $FORM{'comment_file'} in PROCESS_VOTE</strong>\n";
	    }
	}
    }
}

##################
# PRINT_COMMENTS #
##################
sub print_comments
{
#
# takes no arguments
#
# formats and prints all comments
#
# always returns true
#
    print "<h3>$FORM{'comment_title'}</h3>\n";
    print <<END;
<table border=0 bgcolor="$FORM{'comment_body_bgcolor'}" background="$FORM{'comment_body_background'}" 
align="center" width="$FORM{'comment_table_width'}">\n
END
    if (safeopen (PRINTCOMMENTS, "$FORM{'comment_file'}", "<"))
    {
	my @lines;
	if ($FORM{'new_comments_first'})
	{
	    @lines = <PRINTCOMMENTS>;
	}
	else
	{
	    @lines = reverse <PRINTCOMMENTS>;
	}
	while ($line = pop @lines)
	{
	    my ($name, $email, $timestamp, $title, $comment) = split (/\^/, $line);
	    print <<END;
<tr><td>
<table bgcolor=$FORM{'comment_title_bgcolor'} background=$FORM{'comment_title_background'} 
border=0 align=center width="$FORM{'comment_table_width'}">
<tr><td>
<font color="$FORM{'comment_title_font'}">
<strong>$title</strong><br>
Posted by <a href="mailto:$email">$name</a> on $timestamp.
</td>
</td></tr></table>
</td></tr>
<tr><td>
<table bgcolor=$FORM{'comment_body_bgcolor'} background=$FORM{'comment_body_background'} 
border=0 align=center width="$FORM{'comment_table_width'}">
<tr><td>
<font color="$FORM{'comment_body_font'}">
$comment
</font>
</td></tr></table>
</td></tr>
END
        }
	safeclose(PRINTCOMMENTS);
	print "</table>\n";
    }
    else
    {
	if ($FORM{'debug'})
	{
	    print "<strong>Unable to open comment file $FORM{'comment_file'} in PRINT_COMMENTS</strong>\n";
	}
    }
    return 1;
}

################
# PRINT_HEADER #
################
sub print_header
{
# Takes no arguments
#
# Prints the contents of HEAD_URL or a canned
# opening if HEAD_URL is undefined or not found
#
# Always returns 1
#
    unless (http_dump($FORM{'head_url'}))
    {
	print <<END;
<html><head><title>$FORM{'page_title'}</title></head>
<body bgcolor="$FORM{'bgcolor'}" background="$FORM{'background'}" text="$FORM{'text_color'}" 
alink="$FORM{'alink_color'}" vlink="$FORM{'vlink_color'}" link="$FORM{'link_color'}"><center>
END
    }
    return 1;
}

################
# PRINT_FOOTER #
################
sub print_footer
{
#
# Takes no arguments
#
# Prints a small blurb about vote followed by 
# the contents of FOOT_URL or a canned closing
# if FOOT_URL is undefined or not found
#
# Always returns 1
#
    print <<END;
<center><font size=1>
<hr width=25%>
Generated by PLVote $version<br>
Written by <a href="mailto:rsw\@phonon.net">Riad Wahby</a><br>
Available at <a href="$ftpurl">$ftpurl</a><br>
</font></center>
END
    unless (http_dump($FORM{'foot_url'}))
    {
	print "</center></body></html>\n";
    }
    return 1;
}

#############
# HTTP_DUMP #
#############
sub http_dump
{
#
# This sub takes one argument, the URL of the
# HTML to "dump".
#
# Prints the contents of its argument
#
# On success, returns 1.
# Otherwise, it returns 0.
#

    my $url = $_[0];
    unless ($url)
    {
	if ($FORM{'debug'})
	{
	    print "<strong>No URL specified in HTTP_DUMP</strong><br>\n";
	}
	return 0;
    }
    my $ua = LWP::UserAgent->new();
    my $request = HTTP::Request->new(GET => "$url");
    my $response = $ua->request($request);
    if ($response->is_success)
    {
	print $response->content;
	return 1;
    }
    else
    {
	if ($FORM{'debug'})
	{
	    print $response->error_as_HTML();
	}
	return 0;
    }
}

#########
# TOUCH #
#########
sub touch
{
#
# takes one argument, the file to touch
#
# basically touch from UNIX
#
# returns 1 on success, 0 otherwise
#
    if (-e $_[0])
    {
	return 0;
    }
    else
    {
	if (open (TOUCH, ">$_[0]"))
	{
	    close(TOUCH);
	    return 1;
	}
	else
	{
	    return 0;
	}
    }
}

############
# SAFEOPEN #
############
sub safeopen
{
#
# takes three arguments, 
# file descriptor, filename, 
# and i/o direction (i.e. >, >>, or <)
#
# does basic file locking (used for 
# portability to NT systems)
#
# does not return until file is opened
#
    my $tries = 0;
    until (touch "$_[1].lock")
    {
	$tries++;
	if ($tries == $maxtries)
	{
	    return 0;
	    if ($FORM{'debug'})
	    {
		print "Unable to touch $_[1].lock in SAFEOPEN<br>\n";
	    }
	}
    }
    open ($_[0], "$_[2]$_[1]");
    $OPENFILES{$_[0]} = $_[1];
    return 1;
}

#############
# SAFECLOSE #
#############
sub safeclose
{
#
# takes one argument, the filehandle to close
#
# implements closing of locked files
#
# 
#
    close $_[0];
    unlink ($OPENFILES{$_[0]} . ".lock");
}

##############
# PRINT_HELP #
##############
sub print_help
{
    print <<END;
<html><head><title>PLVote $version</title></head>
<body bgcolor="#FFFFFF" text="#000000">
<h2>PLVote $version HTML Help</h2>
<p>If you you are unsure why you are getting this help message,
it is probably because you did not provide all required arguments.
Please check your setup.

<h3>General Use</h3> 

<p>To use this script, all that is necessary is that arguments be
POSTed to it using the HTML &lt;form&gt; tag.  For example,
<pre>
&lt;form method=POST action=http://foo.com/cgi-bin/vote.pl&gt;
&lt;input type=hidden name=VOTE_FILE value="d:\\full\\path\\vote"&gt;
&lt;input type=hidden name=COMMENT_FILE value="d:\\full\\path\\comment"&gt;
&lt;input type=hidden name=BARSRC value="http://foo.com/image/bar.gif"&gt;
&lt;input type=radio name=VOTE_QNUM value=1&gt;This is answer 1
&lt;input type=hidden name=A1 value="This is answer 1"&gt;
&lt;input type=radio name=VOTE_QNUM value=2&gt;This is answer 2
&lt;input type=radio name=A2 value="This is answer 2"&gt;
&lt;input type=submit value="Vote"&gt;
&lt;input type=reset value="Reset"&gt;
&lt;/form&gt;
</pre>

<h3>Required Arguments</h3>
<p>The following are required.  If you do not specify them, you
will get this message.
<ul>
<li><pre>VOTE_FILE</pre> - the path on the server to the file in which
vote data should be stored.  The directory must exist, but if the file
does not it will be created.<br><br>

<li><pre>COMMENT_FILE</pre> - the path on the server to the file in
which comment data should be stored.  The directory must exist, but if
the file does not it will be created.<br><br>

<li><pre>BARSRC</pre> - the URL of the image to use as the bar for a
bar graph.  It will be stretched a bit, so a solid colored small
graphic (such as a 25x25 px square) is recommended.<br><br>
</ul>

<h3>Optional Arguments</h3>
<p>The following are optional.  They allow you finer control over the
appearance of the output.
<ul>
<li><pre>HEAD_URL</pre> - the URL to a page to display above the
PLVote output.  This file must contain at least the following:
    <pre>&lt;html&gt;&lt;body&gt;</pre><br><br>

<li><pre>FOOT_URL</pre> - the URL to a page to display below the
PLVote output.  This file must contain at least the following:
    <pre>&lt;/body&gt;&lt;/html&gt;</pre><br><br>

<li><pre>PAGE_TITLE</pre> - the title for the page to display.
Ignored if HEAD_URL is specified.<br><br>

<li><pre>VOTE_TITLE</pre> - the title to display above the voting
results table.  Default is "Vote Results".<br><br>

<li><pre>COMMENT_TITLE</pre> - the title to display above the comments
table.  Default is "Comments".<br><br>

<li><pre>BAR_HEIGHT</pre> - the height in pixels of the bar for the
graph.  Default is 25.<br><br>

<li><pre>BGCOLOR</pre> - the background color for the output and all
tables except those otherwise specified.  Specify as "#RRGGBB".  Default
is "#FFFFFF" (white).  Ignored if HEAD_URL is specified.<br><br>

<li><pre>BACKGROUND</pre> - the background image for the output and
all tables except those otherwise specified.  Specify URL of image.
Default is none.  Ignored if HEAD_URL is specified.<br><br>

<li><pre>GRAPH_WIDTH</pre> - the width of the table cell containing the
voting bar graph.  Default is 390.<br><br>

<li><pre>VOTE_BGCOLOR</pre> - the background color for the table of
voting stats.  Default is the same as specified for BGCOLOR.<br><br>

<li><pre>VOTE_BACKGROUND</pre> - the background image for the table of
voting stats.  Default is the same as specified for
BACKGROUND.<br><br>

<li><pre>COMMENT_TITLE_FONT</pre> - the font color to use in the
comment titles.  Default is the same as specified for
COMMENT_BODY_FONT.<br><br>

<li><pre>COMMENT_BODY_FONT</pre> - the font color to use in the body
of a comment.  Default is "#000000" (black).<br><br>

<li><pre>COMMENT_TABLE_WIDTH</pre> - the width of the comment table in
pixels or percent.  Default is 100%.<br><br> 

<li><pre>COMMENT_TITLE_BGCOLOR</pre> - the background color for the
titles of the comment table.  Default is "#00B4ED" (a light
blue).<br><br>

<li><pre>COMMENT_TITLE_BACKGROUND</pre> - the background image for the
titles of the comment table.  Default is the same as specified for
BACKGROUND.<br><br>

<li><pre>COMMENT_BODY_BGCOLOR</pre> - the background color for the
comment bodies of the comment table.  Default is the same as specified
for BGCOLOR.<br><br>

<li><pre>COMMENT_BODY_BACKGROUND</pre> - the background color for the
comment bodies of the comment table.  Default is the same as specified
for BACKGROUND.<br><br>

<li><pre>TEXT_COLOR</pre> - the text color for the output.  Specify as
"#RRGGBB".  Ignored if HEAD_URL is specified.  (Set text colors in the
<tt>&lt;body&gt;</tt> tag)<br><br>

<li><pre>LINK_COLOR</pre> - the link color for the output.  Ignored if
HEAD_URL is specified.<br><br>

<li><pre>VLINK_COLOR</pre> - the visited link color for the
output.  Ignored if HEAD_URL is specified.<br><br>

<li><pre>ALINK_COLOR</pre> - the activated link color for the
output.  Ignored if HEAD_URL is specified.<br><br>

<li><pre>NEW_COMMENTS_FIRST</pre> - if this is set to "1" all comments
will be added to the beginning of the comment file instead of the end.
A value of "0" disables.  Default is 0.<br><br>

<li><pre>SHOW_COMMENTS</pre> - if this is set to "1" all comments will
be formatted and shown.  A setting of "0" will disable the comments
display.  Default is 0.<br><br>

<li><pre>DEBUG</pre> - if this is set to "1" debug messages will be
printed in the HTML (such as unable to open file, etc).  A setting of
"0" disables.  It is recommended that debug is used until the vote
script is working.  It should be turned off for production pages.<br><br>

</ul>

<h3>Voting</h3>
<p>To actually place a vote, the following fields should be filled in.
All except the first two are optional.  If VOTE_QNUM or A(VOTE_QNUM)
is blank, results will be printed with no modification to the
statistics.
<ul>
<li><pre>VOTE_QNUM</pre> - the option the user has chosen.  This
should be done as a set of radio buttons so that only one VOTE_QNUM
can be specified.<br><br>

<li><pre>Ax</pre> - The text of choice x.  For instance, A1 might be
Yes and A2 might be No.  This is stored along with the vote data, and
is updated every time the script is run, which means you can change
the wording of a question and it will be changed in the vote data.<br><br>

<li><pre>NAME</pre> - the name of the user, stored along with the
comment.  If blank, it will be changed to Anonymous.<br><br>

<li><pre>EMAIL</pre> - the email address of the user, stored along
with the comment.<br><br>

<li><pre>TITLE</pre> - the title of the comment.  If blank, it will be
changed to "Untitled Comment".<br><br>

<li><pre>COMMENT</pre> - the text of the comment.  It can be from any
input, but a text box or textarea are the best method of getting the
comment.<br><br>
</ul>
END
    $FORM{'foot_url'} = "";
    print_footer();
}
